@@ -7,12 +7,22 @@ use either::Either;
 use regex::Regex;
 #[cfg(feature = "regex")]
 use std::borrow::Borrow;
+use std::borrow::Cow;
 use std::env;
 use std::ffi::OsStr;
-#[cfg(feature = "regex")]
+#[cfg(any(feature = "regex", target_os = "windows"))]
 use std::fs;
 use std::iter;
-use std::path::{Path, PathBuf};
+use std::path::{Component, Path, PathBuf};
+
+// Home dir shim, use home crate when possible. Otherwise, return None
+#[cfg(any(windows, unix, target_os = "redox"))]
+use home::home_dir;
+
+#[cfg(not(any(windows, unix, target_os = "redox")))]
+fn home_dir() -> Option<std::path::PathBuf> {
+    None
+}
 
 pub trait Checker {
     fn is_valid(&self, path: &Path) -> bool;
@@ -73,14 +83,20 @@ impl Finder {
             }
             _ => {
                 // Search binary in PATHs(defined in environment variable).
-                let p = paths.ok_or(Error::CannotFindBinaryPath)?;
-                let paths: Vec<_> = env::split_paths(&p).collect();
+                let paths =
+                    env::split_paths(&paths.ok_or(Error::CannotGetCurrentDirAndPathListEmpty)?)
+                        .collect::<Vec<_>>();
+                if paths.is_empty() {
+                    return Err(Error::CannotGetCurrentDirAndPathListEmpty);
+                }
 
                 Either::Right(Self::path_search_candidates(path, paths).into_iter())
             }
         };
 
-        Ok(binary_path_candidates.filter(move |p| binary_checker.is_valid(p)))
+        Ok(binary_path_candidates
+            .filter(move |p| binary_checker.is_valid(p))
+            .map(correct_casing))
     }
 
     #[cfg(feature = "regex")]
@@ -93,7 +109,10 @@ impl Finder {
     where
         T: AsRef<OsStr>,
     {
-        let p = paths.ok_or(Error::CannotFindBinaryPath)?;
+        let p = paths.ok_or(Error::CannotGetCurrentDirAndPathListEmpty)?;
+        // Collect needs to happen in order to not have to
+        // change the API to borrow on `paths`.
+        #[allow(clippy::needless_collect)]
         let paths: Vec<_> = env::split_paths(&p).collect();
 
         let matching_re = paths
@@ -130,12 +149,14 @@ impl Finder {
     where
         P: IntoIterator<Item = PathBuf>,
     {
-        let new_paths = paths.into_iter().map(move |p| p.join(binary_name.clone()));
+        let new_paths = paths
+            .into_iter()
+            .map(move |p| tilde_expansion(&p).join(binary_name.clone()));
 
         Self::append_extension(new_paths)
     }
 
-    #[cfg(unix)]
+    #[cfg(not(windows))]
     fn append_extension<P>(paths: P) -> impl IntoIterator<Item = PathBuf>
     where
         P: IntoIterator<Item = PathBuf>,
@@ -148,29 +169,31 @@ impl Finder {
     where
         P: IntoIterator<Item = PathBuf>,
     {
+        use once_cell::sync::Lazy;
+
         // Sample %PATHEXT%: .COM;.EXE;.BAT;.CMD;.VBS;.VBE;.JS;.JSE;.WSF;.WSH;.MSC
         // PATH_EXTENSIONS is then [".COM", ".EXE", ".BAT", …].
         // (In one use of PATH_EXTENSIONS we skip the dot, but in the other we need it;
         // hence its retention.)
-        lazy_static! {
-            static ref PATH_EXTENSIONS: Vec<String> =
-                env::var("PATHEXT")
-                    .map(|pathext| {
-                        pathext.split(';')
-                            .filter_map(|s| {
-                                if s.as_bytes().first() == Some(&b'.') {
-                                    Some(s.to_owned())
-                                } else {
-                                    // Invalid segment; just ignore it.
-                                    None
-                                }
-                            })
-                            .collect()
-                    })
-                    // PATHEXT not being set or not being a proper Unicode string is exceedingly
-                    // improbable and would probably break Windows badly. Still, don't crash:
-                    .unwrap_or(vec![]);
-        }
+        static PATH_EXTENSIONS: Lazy<Vec<String>> = Lazy::new(|| {
+            env::var("PATHEXT")
+                .map(|pathext| {
+                    pathext
+                        .split(';')
+                        .filter_map(|s| {
+                            if s.as_bytes().first() == Some(&b'.') {
+                                Some(s.to_owned())
+                            } else {
+                                // Invalid segment; just ignore it.
+                                None
+                            }
+                        })
+                        .collect()
+                })
+                // PATHEXT not being set or not being a proper Unicode string is exceedingly
+                // improbable and would probably break Windows badly. Still, don't crash:
+                .unwrap_or_default()
+        });
 
         paths
             .into_iter()
@@ -180,19 +203,58 @@ impl Finder {
                     Box::new(iter::once(p))
                 } else {
                     // Appended paths with windows executable extensions.
-                    // e.g. path `c:/windows/bin` will expend to:
-                    // c:/windows/bin.COM
-                    // c:/windows/bin.EXE
-                    // c:/windows/bin.CMD
+                    // e.g. path `c:/windows/bin[.ext]` will expand to:
+                    // [c:/windows/bin.ext]
+                    // c:/windows/bin[.ext].COM
+                    // c:/windows/bin[.ext].EXE
+                    // c:/windows/bin[.ext].CMD
                     // ...
-                    Box::new(PATH_EXTENSIONS.iter().map(move |e| {
-                        // Append the extension.
-                        let mut p = p.clone().into_os_string();
-                        p.push(e);
+                    Box::new(
+                        iter::once(p.clone()).chain(PATH_EXTENSIONS.iter().map(move |e| {
+                            // Append the extension.
+                            let mut p = p.clone().into_os_string();
+                            p.push(e);
 
-                        PathBuf::from(p)
-                    }))
+                            PathBuf::from(p)
+                        })),
+                    )
                 }
             })
     }
 }
+
+fn tilde_expansion(p: &PathBuf) -> Cow<'_, PathBuf> {
+    let mut component_iter = p.components();
+    if let Some(Component::Normal(o)) = component_iter.next() {
+        if o == "~" {
+            let mut new_path = home_dir().unwrap_or_default();
+            new_path.extend(component_iter);
+            Cow::Owned(new_path)
+        } else {
+            Cow::Borrowed(p)
+        }
+    } else {
+        Cow::Borrowed(p)
+    }
+}
+
+#[cfg(target_os = "windows")]
+fn correct_casing(mut p: PathBuf) -> PathBuf {
+    if let (Some(parent), Some(file_name)) = (p.parent(), p.file_name()) {
+        if let Ok(iter) = fs::read_dir(parent) {
+            for e in iter.filter_map(std::result::Result::ok) {
+                if e.file_name().eq_ignore_ascii_case(file_name) {
+                    p.pop();
+                    p.push(e.file_name());
+                    break;
+                }
+            }
+        }
+    }
+    p
+}
+
+#[cfg(not(target_os = "windows"))]
+fn correct_casing(p: PathBuf) -> PathBuf {
+    p
+}
